"""
This code is modified from the tinytag project:

https://github.com/devsnd/tinytag

Original license below:

MIT License

Copyright (c) 2014-2017 Tom Wallroth

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import struct
import os
import codecs

def parse_pages(fh):
    # for the spec, see: https://wiki.xiph.org/Ogg
    previous_page = b''  # contains data from previous (continuing) pages
    header_data = fh.read(27)  # read ogg page header
    while len(header_data) != 0:
        header = struct.unpack('<4sBBqIIiB', header_data)
        oggs, version, flags, pos, serial, pageseq, crc, segments = header
        # self._max_samplenum = max(self._max_samplenum, pos)
        if oggs != b'OggS' or version != 0:
            raise Exception('Not a valid ogg file!')
        segsizes = struct.unpack('B'*segments, fh.read(segments))
        total = 0
        for segsize in segsizes:  # read all segments
            total += segsize
            if total < 255:  # less than 255 bytes means end of page
                yield previous_page + fh.read(total)
                previous_page = b''
                total = 0
        if total != 0:
            if total % 255 == 0:
                previous_page += fh.read(total)
            else:
                yield previous_page + fh.read(total)
                previous_page = b''
        header_data = fh.read(27)

def parse_comment(fh):
    # for the spec, see: http://xiph.org/vorbis/doc/v-comment.html
    # discnumber tag based on: https://en.wikipedia.org/wiki/Vorbis_comment
    # https://sno.phy.queensu.ca/~phil/exiftool/TagNames/Vorbis.html
    comment_type_to_attr_mapping = {
        'album': 'album',
        'albumartist': 'albumartist',
        'title': 'title',
        'artist': 'artist',
        'date': 'year',
        'tracknumber': 'track',
        'discnumber': 'disc',
        'genre': 'genre',
        'description': 'comment',
        'composer': 'composer',
    }
    vendor_length = struct.unpack('I', fh.read(4))[0]
    fh.seek(vendor_length, os.SEEK_CUR)  # jump over vendor
    elements = struct.unpack('I', fh.read(4))[0]

    metadata = []

    for i in range(elements):
        length = struct.unpack('I', fh.read(4))[0]
        try:
            keyvalpair = codecs.decode(fh.read(length), 'UTF-8')
        except UnicodeDecodeError:
            continue
        if '=' in keyvalpair:
            key, value = keyvalpair.split('=', 1)
            fieldname = comment_type_to_attr_mapping.get(key.lower())
            if fieldname:
                metadata.append((fieldname, value))

    return metadata



"""
Contains a command-line interface implemented with argparse.
"""

from traktor_nowplaying.core import Listener
from traktor_nowplaying.options import PORT, QUIET, OUTPUT_FORMAT, INTERACTIVE, APPEND, MAX_TRACKS
from traktor_nowplaying.version import __version__
import argparse
import signal
import sys


DESCRIPTION = 'Use Traktor\'s broadcast functionality to extract metadata about the currently playing song'
EPILOG = f'Note that you must configure Traktor to broadcast to localhost and the port specified with the -p, or --port option (defaults to {PORT}). For the format setting you can use anything, but I recommend choosing the lowest bitrate for the sample rate of your system, so most commonly the best choice is 44100 Hz, 64 Kbps.'
PROGRAM_NAME = 'traktor_nowplaying'

parser = argparse.ArgumentParser(
    description=DESCRIPTION,
    epilog=EPILOG,
    prog=PROGRAM_NAME,
)

parser.add_argument('-p', '--port', default=PORT,
    type=int,
    help='Port to listen on for broadcasts from Traktor'    
)

parser.add_argument('-q', '--quiet', default=QUIET,
    action='store_true',
    help='Suppress console output of currently playing song'
)

parser.add_argument('-f', '--format', default=OUTPUT_FORMAT,
    help='Custom format to use when outputting tracks'
)

parser.add_argument('-o', '--outfile', default=None,
    help='Provide a file path to which the currently playing song should be written',
)

parser.add_argument('-t', '--template', default=None,
    help='Template file to use for output. Templating is implemented using Bottle SimpleTemplate (https://bottlepy.org/docs/0.12/stpl.html). See README for more details on use. Note: the --format options is ignored when using a custom template file. Take care when using templates provided by others on the internet as they can contain malicious code.'
)

parser.add_argument('-a', '--append', default=APPEND,
    action='store_true',
    help='If writing to file, appends newest track to end of file instead of overwriting the file'
)

parser.add_argument('-m', '--max-tracks', default=MAX_TRACKS,
    action='store',
    type=int,
    help='If appending to a file, the maximum number of tracks to keep in file (by default there is no limit)'
)

parser.add_argument('-i', '--interactive', default=INTERACTIVE,
    action='store_true',
    help='Interactive mode allows for settings to be specified at runtime. These override command line options.'
)

parser.add_argument('-v', '--version',
    action='version',
    version='%(prog)s {version}'.format(version=__version__)
)


def _exit(status=None):
    print('\n Exiting')
    sys.exit(status)

# capture Ctrl + C
signal.signal(signal.SIGINT, lambda signum, frame: _exit())

def _get_port_interactively():
    MIN_PORT = 1
    MAX_PORT = 65535

    port = None

    while True:
        try:
            port = input(f'Please specify port to listen on (press Enter to accept default: {PORT}): ').strip()

            # accepted default
            if port == '':
                break
            
            # attempt to convert to int
            port = int(port)

            # and make sure it is valid
            if port < MIN_PORT or port > MAX_PORT:
                raise ValueError
        except ValueError:
            print(f'Port needs to be an integer between {MIN_PORT}-{MAX_PORT}')
        except KeyboardInterrupt:
            _exit()

    return port

def interactive():
    print('Interactive mode is active. Press Enter to accept the default for each choice.')
    port = _get_port_interactively()
    # Ask for y/n but also strip extra space
    quiet = input(f'Quiet mode (suppresses console output)? (Y/[N], default N): ').lower().strip() or 'n'
    # and also accept "yes" or really yXX by truncating string
    quiet = (quiet[0] == 'y')
    outfile = input('Path of file to output current playing song to (no file output by default): ').strip()
    _args = []

    if port:
        _args.extend(['--port', port])
    if quiet:
        _args.extend(['--quiet'])
    if outfile:
        _args.extend(['--outfile', outfile])

        append = input('Append latest track to end of file? (Y/[N], default N): ').lower().strip() or 'n'
        # and also accept "yes" or really yXX by truncating string
        append = (append[0] == 'y')
        if append:
            _args.extend(['--append'])
        
            try:
                max_tracks = input('Number of tracks to keep in file (default no limit): ')
                if int(max_tracks) > 0:
                    _args.extend(['--max-tracks', max_tracks])
            except ValueError:
                pass

    print(f"{parser.prog} {' '.join(_args)}")

    return _args

def want_interactive():
    while True:
        answer = input(' '.join([
            'Type the letter i (then press Enter) to set options',
            'interactively or press h (then press Enter) for help, '
            'or just press Enter to continue with default options: '
        ])).lower().strip()

        if answer == 'h':
            parser.print_help()
            print('\n')
            continue

        return answer == 'i'

def _read_template_file(path):
    try:
        with open(path, 'r') as f:
            return f.read()
    except:
        print(f'Error encountered while trying to read template file ({path}).')

    return

def main():
    args = parser.parse_args()
    # set arguments interactively if interactive flag is passed
    # or if the user passes no arguments and desires interactive mode
    if args.interactive or (len(sys.argv) == 1 and want_interactive()):
        args = parser.parse_args(interactive())

    template = _read_template_file(args.template) if args.template else None

    listener = Listener(
        port=args.port,
        quiet=args.quiet,
        output_format=args.format,
        outfile=args.outfile,
        template=template,
        append=args.append,
        max_tracks=args.max_tracks
    )
    listener.start()

if __name__ == '__main__':
    main()



from .options import PORT, QUIET, OUTPUT_FORMAT, APPEND, MAX_TRACKS
from .ogg import parse_comment, parse_pages
from .bottle import SimpleTemplate, TemplateError
from collections import deque
import functools
import http.server
import html
import socketserver
import pathlib
import types
import io
import os


def create_request_handler(callbacks):
    """Creates an HTTP request handler with a custom callback"""

    class TraktorHandler(http.server.BaseHTTPRequestHandler):
        """Simpler handler for Traktor requests."""

        def do_SOURCE(self):
            """
            Implement handler for SOURCE requests which Traktor and older
            icecast source cilents send data via a special SOURCE verb.
            """
            # we send response
            self.send_response(200)
            # and headers
            self.end_headers()

            # and now the streaming begins
            for packet in parse_pages(self.rfile):
                walker = io.BytesIO(packet)
                if packet[:7] == b"\x03vorbis":
                    walker.seek(7, os.SEEK_CUR)  # jump over header name
                    metadata = parse_comment(walker)

                    for callback in callbacks:
                        callback(metadata)

        def log_request(self, code='-', size='-'):
            """Do not log messages about HTTP requests."""
            pass

        def log_error(self, format, *args):
            """Do not log messages about HTTP requests."""
            pass

    return TraktorHandler


class TrackWriter:
    """Writes tracks to standard output and/or file."""
    def __init__(self, quiet=QUIET, output_format=OUTPUT_FORMAT, outfile=None, template=None, append=APPEND, max_tracks=MAX_TRACKS):
        self.quiet = quiet
        self.outfile = outfile
        self.append = append

        self.max_tracks = max_tracks if append else 1
        self.tracks = deque(maxlen=self.max_tracks)

        try:
            self.output_format = SimpleTemplate(output_format)
        except TemplateError:
            print('Error encountered while trying to parse output format.')
            self.output_format = SimpleTemplate(OUTPUT_FORMAT)

        self.template = None
    
        if template is not None:
            try:
                self.template = SimpleTemplate(template)
            except TemplateError:
                print('Error encountered while trying to parse template.')

        if self.outfile:
            try:
                self._create_outfile()
            except:
                print(f'Error encountered while trying to write to {self.outfile}.')

    def update(self, data):
        info = dict(data)

        if not ('artist' in info or 'title' in info):
            return

        self.tracks.append(info)

        if not self.quiet:
            self._to_stdout()
        if self.outfile:
            self._to_file()

    def _get_track_string(self, info):
        return html.unescape(self.output_format.render(
            artist=info.get('artist', ''),
            title=info.get('title', '')
        ))

    def _to_stdout(self):
        track_string = self._get_track_string(self.tracks[-1])
        if track_string:
            print(track_string)

    def _create_outfile(self):
        outpath = pathlib.Path(self.outfile)
        outpath.parent.mkdir(parents=True, exist_ok=True)
        outpath.touch(exist_ok=True)

        if not outpath.is_file():
            print(f'{self.outfile} is a directory!')
            raise IsADirectoryError

    def _to_file(self):
        if self.template:
            tracklist = self.template.render(tracks=self.tracks)
        else:
            tracklist = '\n'.join(self._get_track_string(t) for t in self.tracks)

        with open(self.outfile, 'w', encoding='utf-8') as f:
            f.write(tracklist)


class Listener():
    """Listens to Traktor broadcast, given a port."""

    def __init__(self, port=PORT, quiet=QUIET, output_format=OUTPUT_FORMAT, outfile=None, template=None, append=APPEND, max_tracks=MAX_TRACKS, custom_callback=None):
        self.port = port
        self.quiet = quiet
        self.output_format = output_format
        self.outfile = outfile
        self.template = template
        self.append = append
        self.max_tracks = max_tracks
        self.custom_callback = custom_callback

    def start(self):
        """Start listening to Traktor broadcast."""

        callbacks = []

        if self.outfile is not None or not self.quiet:
            writer = TrackWriter(
                quiet=self.quiet,
                output_format=self.output_format,
                outfile=self.outfile,
                template=self.template,
                append=self.append,
                max_tracks=self.max_tracks
            )
            callbacks.append(writer.update)

        if not self.quiet:
            print(f'Listening on port {self.port}.')
            if self.outfile:
                print(f'Outputting to {self.outfile}')

        if self.custom_callback:
            callbacks.append(self.custom_callback)

        # create a request handler with appropriate callback
        handler = create_request_handler(callbacks=callbacks)

        with socketserver.TCPServer(('', self.port), handler) as httpd:
            try:
                httpd.serve_forever()
            except KeyboardInterrupt:
                httpd.server_close()



"""Setup script for traktor_nowplaying."""
from setuptools import setup, find_packages, Command
from shutil import rmtree
import codecs
import io
import os
import sys

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    """Return multiple read calls to different readable objects as a single string."""
    return codecs.open(os.path.join(here, *parts), 'r').read()

NAME = 'traktor_nowplaying'
DESCRIPTION = f'{NAME} uses Traktor\'s broadcast functionality to extract metadata about the currently playing song.'
LONG_DESCRIPTION = read('README.md')
URL = 'https://github.com/radusuciu/traktor_nowplaying'
EMAIL = 'radusuciu@gmail.com'
AUTHOR = 'Radu Suciu'

# not doing import because do not want to have to load module
# before it has been installed
version_path = os.path.join(here, 'traktor_nowplaying/version.py')
exec(read(version_path))
VERSION = __version__


# from github.com/kennethreitz/setup.py
class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    url=URL,
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    python_requires='>=3.6.2',
    packages=find_packages(exclude=('tests',)),
    entry_points={
        'console_scripts': ['traktor_nowplaying=traktor_nowplaying.cli:main']
    },
    include_package_data=True,
    platforms='any',
    zip_safe=True,
    license='MIT License',
    classifiers=[
        'Environment :: Console',
        'Topic :: Multimedia :: Sound/Audio',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)